"""This is a module supporting autiting of EDG users.

Provides  routines to allow mining of various useful bits of information from
the logging files produced by the EDG and globus middleware and the gridmap
database.
Author: t.s.eves@rl.ac.uk
Version: 1.0
"""
import os, re, string, sys, time

MKGRIDMAP_CFG_DEFAULT_PATH = '/opt/edg/etc/mkgridmap.conf'
GATEKEEPER_LOG_DEFAULT_PATH = '/var/log/globus-gatekeeper.log'
GRIDMAPDIR_DEFAULT_PATH = '/etc/grid-security/gridmapdir'



def user2VO(user_name, config_path = MKGRIDMAP_CFG_DEFAULT_PATH) :
	'''Return the list of VOs that the pool account belongs to.'''
	# Match everything but the trailing digits of the account name.
	prefix_match = re.match(r"([\w\D]*\D)\d+", user_name)

	if not prefix_match:		# if no prefix is found raise an exception.
		raise ValueError, "not a pool account name: '" + user_name +"'"

	# Precompile a regular expression to match group definition lines and
	# extract the VO field.
	vo_rec = re.compile(r"group\s+(\S+)\s+\." + prefix_match.group(1))

	# find lines matching the compiled regex 'vo_rec'
	results = []
	cfg_file = open(config_path, "r", 1)
	while 1:
		lines = cfg_file.readlines(8*1024)
		if not lines: break
		for line in lines:
			vo_match = vo_rec.match(line)
			if vo_match :
				results.append(vo_match.group(1))

	return results



def read_gk_events(log_path=None) :
	return GKESeq(log_path \
		or os.environ.get('GATEKEEPER_LOG', GATEKEEPER_LOG_DEFAULT_PATH))



def startswith(ds, ts) :
	'''Returns true if ds begins with ts.'''
	return ds[:len(ts)] == ts


def tail(line, sep) :
	cut_pt = string.find(line, sep)
	return cut_pt != -1 and line[cut_pt+len(sep):] or None



def rtail(line, sep) :
	cut_pt = string.rfind(line, sep)
	return cut_pt != -1 and line[cut_pt+len(sep):] or None



# Definition of the gatekeeper events sequence object class and associated objects
# Set of posible events.
class GKESeq :
	'''Represents a globus gatekeeper log file as a sequence of entries.

Parse a gatekeeper log file presenting it as a seqeunce of events,
though it reads each event on demand and does not hold the file in memory.

Events consist of three types, starting, stopping and jobs returned as a
3-tuple consisiting of:
    (HEADER, TYPE, BODY)
where HEADER is a 2-tuple:
    (TIMESTAMP - seconds since epoch, IP address of requester)
        Either of these may be None if the information is not suplied.
TYPE is one of GK_EVT_<START|STOP|JOB>, and BODY is formated
depending on the TYPE and wether or not it failed.
For the following type the elements wil be:

starting:
HEADER = (TIMESTAMP, None) - event originated localy.
TYPE = "starting"
BODY = (PID of the gatekeeper daemon, path of executable)

stopping:
HEADER = (None, None) - no time stamp recorded, event originated localy.
TYPE = "stopping"
BODY = signal number received.

job request:
HEADER = (TIMESTAMP, IP)
TYPE = "job"
BODY =
success     : (DN, service requested, usr_nm, uid, gid, (path to child, PID of child))
auth-failure: ("Failure", failure message string)
job-failure: ("Failure", failure message string, DN)

OBJECT DATA

eventno
	The index of the next event.
lineno
	The line number of the next line to be read.
path
	The path of the log file being parsed.
'''
	__Notice6 = intern('Notice: 6: ')
	__Stopping = intern('Notice: 5: Gatekeeper received signal:')
	__Notice5 = intern('Notice: 5: ')
	__Notice0 = intern('Notice: 0: ')
	__DiscardThreshold = len(__Notice0)
	__ChildPidRegEx = re.compile(r"\w+ (\d+) \w+")


	def __init__(self, log_path) :
		'''Opens the log path and initialise an GKESeq object.'''
		self.eventno = 0
		self.__log_file = None
		self.__log_file = open(log_path, "r", 1024*1024)
		self.path = log_path
		self.lineno = 1
		self.__ln = 0
		self.__linebuf = []


	def __del__(self) :
		'''Closes the log file.'''
		if self.__log_file :
			self.__log_file.close()


	def __get_next_line(self) :
		'''Retrieves the next non-blank line'''
		if self.__ln == len(self.__linebuf):
			self.__linebuf = self.__log_file.readlines(256*1024)
			if not self.__linebuf : return ''
			self.__ln = 0
		line = self.__linebuf[self.__ln]
		self.__ln = self.__ln + 1
		self.lineno = self.lineno + 1
		while len(line) < 16 :
			if self.__ln == len(self.__linebuf):
				self.__linebuf = self.__log_file.readlines(256*1024)
				if not self.__linebuf : return ''
				self.__ln = 0
			line = self.__linebuf[self.__ln]
			self.__ln = self.__ln + 1
			self.lineno = self.lineno + 1
		return line


	def __getitem__(self, idx) :
		'''Returns the event at or imediately after the current file position.

Returns either a 3-tuple describing the event or None if the log entry was
unreadable for any reason.
'''
		if self.eventno != idx :		# Check the rules are being followed.
			print self.eventno, idx
			raise RuntimeError, "GKESeq object accessed out of order"

		entry = None
		# Find the start of an event.
		try :
			line = self.__get_next_line()
			while line :
				# Found a job log entry or a starting entry.
				if startswith(line, GKESeq.__Notice6) :
					line = line[len(GKESeq.__Notice6):]
					if startswith(line, 'Got ') : # Found a job event
						cmps = string.split(line[len('Got connection '):], ' at ', 1)
						entry = self.__read_job_entry()
						entry = entry and \
							((time.mktime(time.strptime(cmps[1])[:-1]+(-1,)), cmps[0]), \
								GK_EVT_JOB, entry) or None
						break
					else :
						cmps = string.split(line, ' pid=', 1)
						if len(cmps) == 2 :	# Found starting event
							job = cmps[0]
							cmps = string.split(cmps[1], ' starting at ',1)
							entry = (time.mktime(time.strptime(cmps[1])[:-1]+(-1,)), None),\
								GK_EVT_START,\
								(long(cmps[0]), job)
						break
				elif startswith(line, GKESeq.__Stopping) :
					entry = (None, None), GK_EVT_STOP, \
						int(line[len(GKESeq.__Stopping):])
					break;

				line = self.__get_next_line()
		except EnvironmentError:
			entry = None
			pass
		if not line :		# Have we reached the end of the file
			raise IndexError
		self.eventno = self.eventno + 1
		return entry


	def __find_failure(self) :
		'''Searches forward in the file for a Failure entry or a Notice entry.

Returns either a 2-tuple if a failure is found or an 0-tuple if no failure
exists and a Notice etnry was found instead
'''
		self.__read_and_verify(lambda l: not (startswith(l, 'Failure: ') or startswith(l, 'Notice: ')))
		lines = self.__read_and_verify(lambda l :startswith(l, 'Failure: '))

		return lines and ('failure', string.split(string.strip(lines[0]), ": ", 1)[1])


	def __read_job_entry(self) :
		'''Parses and returns a job request event entry.'''

		# The second line should be the "Authenticated globus user: <DN>" line.
		dn = tail(self.__get_next_line(), " user: ")
		if not dn : return tuple(self.__find_failure())
		dn = string.strip(dn)

		# Skip the Notice: 0: entry, but check it's correct.
		if not startswith(self.__get_next_line(), GKESeq.__Notice0):
			body = self.__find_failure()
			return body and (body[0], body[1], dn)

		# Now read in the rest of the "Notice: 5" log entrys
		# and grab the remaining Notice 0 info about the child.
		lines = self.__read_and_verify(lambda l: startswith(l, GKESeq.__Notice5)) \
			+ self.__read_and_verify(lambda l: startswith(l, GKESeq.__Notice0))
		if len(lines) != 7 :
			body = self.__find_failure()
			return body and (body[0], body[1], dn)

		try :
			# body == (DN, service, uid, gid, (execed, pid))
			body = dn,								\
			string.strip(rtail(lines[0], ": ")),	\
			string.strip(rtail(lines[1], ": ")),	\
			long(rtail(lines[2], ": ")),			\
			long(rtail(lines[3], ": ")),			\
			(string.strip(tail(lines[4], "ing ")), 	\
				long(string.split(tail(lines[6], 'Child '))[0]))
		except ValueError:
			body = self.__find_failure()
			body = body and (body[0], body[1], dn)
		return body


	def __read_and_verify(self, verify_fn) :
		'''Reads as many lines as can be verified with the fn "verify_fn".'''
		results = []
		candidate_line = self.__get_next_line()
		while verify_fn(candidate_line) :
			results.append(candidate_line)
			candidate_line = self.__get_next_line()
		self.__ln = self.__ln - 1
		self.lineno = self.lineno - 1
		self.__linebuf[self.__ln] = candidate_line
		return results


GK_EVT_START = intern('starting')
GK_EVT_STOP = intern('stopping')
GK_EVT_JOB = intern('job')



def user2current_subject(usr_id, dir_path = None, show_errs = None) :
	'''Returns a job entry associated with a pool account for the current mapping.

Scans the gridmap driectory and discovery the current mapping this
significantly faster for the current case than scanning a large log file.
If the path is provided that will be used otherwise the enviroment variable
'GRDIMAPDIR' is queried and if that fails the standard default location is tried.
Returns:
	(0, DN, usr_nm, uid, gid)
	None	:- if no mapping exists.
'''
	import pwd, re, stat, types

	if type(usr_id) == types.LongType :
		usr_id = pwd.getpwduid(usr_id)[0]
	if not dir_path :
		dir_path = os.environ.get('GRIDMAPDIR',GRIDMAPDIR_DEFAULT_PATH)

	# Get the key (the inode number) and search for another with the same inode
	stat_info = os.stat('%s/%s' % (dir_path, usr_id))
	if stat_info[stat.ST_NLINK] != 2 : return None

	key = stat_info[stat.ST_INO]
	for fn in os.listdir(dir_path) :
		if fn != usr_id :
			stat_info = os.stat('%s/%s' % (dir_path, fn))
			if stat_info[stat.ST_INO] == key :
				# Convert from  URL encoded string into a real string.
				dn = re.sub(r'%([a-fA-F0-9]{2,2})', lambda m:chr(string.atoi(m.group(1), 16)), fn)
				return (0, dn, usr_id, stat_info[stat.ST_UID], stat_info[stat.ST_GID])
	return None




def user2subject(usr_id, time_pt = None, log_path = None, show_errs = 1) :
	'''Returns job entry associated with a pool account at a specific time.

Scans the gatekeeper log to discover which DN was mapped to the pool account
user at or immeadiatedly after the time "time_pt".  If time is not a time 9-tuple
in local time and is None the the current time will be used.  If the log_path is
None the enviroment variable 'GATEKEEPER_LOG' is queried and failling that the standard
default install location tried.
Returns:
    (time stamp, DN, usr_nm, uid, gid)
	None	:- if no mapping exists.
'''
	reader = read_gk_events(log_path)
	# Get the time value into the seconds since the epoch.
	time_pt = time.mktime(time_pt)

	result = None
	for ev in reader :
		if not ev :
			if show_errs:
				sys.stderr.write(os.path.basename(sys.argv[0]) \
					+ ": %s: Malformed log entry %d at line %d\n" \
						% (log_path, reader.eventno-1, reader.lineno-1))
				sys.stderr.flush()
			continue
		if ev[1] == GK_EVT_JOB :		# We have a job entry.
			# If the event is after time_pt the stop.
			if ev[0][0] > time_pt :	break
			# Was it a success and the right uid? Then it could be the one.
			elif len(ev[2]) == 6 and (ev[2][2] == usr_id or ev[2][3] == usr_id):
				result = (ev[0][0], ev[2][0]) + tuple(ev[2][2:5])

	return result
