#!/usr/bin/perl 

use strict;
use File::Basename;
use Getopt::Long;

# Setup the options defaults.
use vars qw(%options);
%options = ();
$options{mancfg} = '/etc/man.config';

# Pull in the edg location default from the sysconfig file. 
if (-e '/etc/sysconfig/edg') {
    open CONF, "</etc/sysconfig/edg";
    while (<CONF>) {
	$options{edgloc} = $1 if (/^\s*EDG_LOCATION\s*=\s*(\S+)\s*/);
    }
    close CONF;
} else {
    $options{edgloc} = '/opt/edg';
}

# Get the specified options.
&GetOptions(\%options, "help", "edgloc=s", "mancfg=s");

# Print the help for the command. 
if ($options{help}) {
    usage();
    exit;
}

# Make more convenient variables. 
my $edgloc = $options{edgloc};
my $mancfg = $options{mancfg};

# Make sure that the configuration file exists. 
die "man configuration file ($mancfg) does not exist" unless (-e $mancfg);

# Make sure that the configuration file is readable.
die "man configuration file ($mancfg) cannot be read" unless (-r $mancfg);

# Read in this file and collect the manpath mappings. 
my %current;
open CONF, "<$mancfg";
while (<CONF>) {
    $current{$1} = $2 if (/^\s*MANPATH_MAP\s+(\S+)\s+(\S+)\s*/);
}
close CONF;

# Determine if new mappings need to be added. 
my %add;

my $key = "$edgloc/bin";
my $value = "$edgloc/share/man";

$add{$key} = $value 
    unless ((defined $current{$key}) && ($current{$key} eq $value));

my $key = "$edgloc/sbin";
my $value = "$edgloc/share/man";

$add{$key} = $value 
    unless ((defined $current{$key}) && ($current{$key} eq $value));

# Only proceed if add contains some values.
my @keys = keys %add;
exit if ($#keys<0);

# Ok, move the old file out of the way. 
my $rc = rename "$mancfg", "$mancfg.old";
die "Cannot rename configuration file ($mancfg -> $mancfg.old).\n" 
    unless ($rc);

# Write the new file adding new mappings just before the first one. 
open OLD, "<$mancfg.old";
open NEW, ">$mancfg";
while (<OLD>) {
    if (/^\s*MANPATH_MAP\s+(\S+)\s+(\S+)\s*/) {
	foreach my $key (keys %add) {
	    print NEW "MANPATH_MAP $key $add{$key}\n";
	    delete $add{$key};
	}
	print NEW $_;
    } else {
	print NEW $_;
    }
}
close NEW;
close OLD;

exit;


# Print out the usage for the command. 
sub usage {

    print <<'EOF'

addEDGManMap.pl [--help] [--edgloc=location] [--mancfg=config]

This command updates the man configuration file with mappings from
EDG_LOCATION/bin and EDG_LOCATION/sbin to EDG_LOCATION/share/man. 

--help             print out this usage statement

--edgloc=location  EDG software location (default=/opt/edg)
--mancfg=config    man configuration file (default=/etc/man.config)

EOF
;

}
